package ex.gt;

import bayonet.math.SpecialFunctions;
import blang.core.BlangExtensions;
import blang.core.ConstantSupplier;
import blang.core.DeboxedName;
import blang.core.Distribution;
import blang.core.DistributionAdaptor;
import blang.core.ForwardSimulator;
import blang.core.IntDistribution;
import blang.core.IntDistributionAdaptor;
import blang.core.IntVar;
import blang.core.LogScaleFactor;
import blang.core.Model;
import blang.core.ModelBuilder;
import blang.core.ModelComponent;
import blang.core.Param;
import blang.core.UnivariateModel;
import blang.core.WritableIntVar;
import blang.inits.Arg;
import blang.inits.DesignatedConstructor;
import blang.types.StaticUtils;
import ca.ubc.stat.blang.StaticJavaUtils;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.function.Supplier;
import org.eclipse.xtext.xbase.lib.ExclusiveRange;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

@SuppressWarnings("all")
public class SerialSampling implements Model, UnivariateModel<IntVar>, ForwardSimulator {
  public static class Builder implements ModelBuilder {
    private boolean fromCommandLine = false;
    
    @Arg
    public IntVar maxRealization;
    
    private boolean maxRealization_initialized = false;
    
    public SerialSampling.Builder setMaxRealization(final IntVar maxRealization) {
      maxRealization_initialized = true;
      this.maxRealization = maxRealization;
      return this;
    }
    
    @Arg
    public IntVar populationSize;
    
    private boolean populationSize_initialized = false;
    
    public SerialSampling.Builder setPopulationSize(final IntVar populationSize) {
      populationSize_initialized = true;
      this.populationSize = populationSize;
      return this;
    }
    
    @Arg
    public IntVar nDraws;
    
    private boolean nDraws_initialized = false;
    
    public SerialSampling.Builder setNDraws(final IntVar nDraws) {
      nDraws_initialized = true;
      this.nDraws = nDraws;
      return this;
    }
    
    public SerialSampling build() {
      // For each optional type, either get the value, or evaluate the ?: expression
      if (!fromCommandLine && !maxRealization_initialized)
        throw new RuntimeException("Not all fields were set in the builder, e.g. missing maxRealization");
      final IntVar __maxRealization = maxRealization;
      if (!fromCommandLine && !populationSize_initialized)
        throw new RuntimeException("Not all fields were set in the builder, e.g. missing populationSize");
      final IntVar __populationSize = populationSize;
      if (!fromCommandLine && !nDraws_initialized)
        throw new RuntimeException("Not all fields were set in the builder, e.g. missing nDraws");
      final IntVar __nDraws = nDraws;
      // Build the instance after boxing params
      return new SerialSampling(
        __maxRealization, 
        new ConstantSupplier(__populationSize), 
        new ConstantSupplier(__nDraws)
      );
    }
  }
  
  @DesignatedConstructor
  public static SerialSampling.Builder builderFromCommandLine() {
    Builder result = new Builder();
    result.fromCommandLine = true;
    return result;
  }
  
  private final IntVar maxRealization;
  
  public IntVar getMaxRealization() {
    return maxRealization;
  }
  
  @Param
  private final Supplier<IntVar> $generated__populationSize;
  
  public IntVar getPopulationSize() {
    return $generated__populationSize.get();
  }
  
  @Param
  private final Supplier<IntVar> $generated__nDraws;
  
  public IntVar getNDraws() {
    return $generated__nDraws.get();
  }
  
  /**
   * Utility main method for posterior inference on this model
   */
  public static void main(final String[] arguments) {
    StaticJavaUtils.callRunner(Builder.class, arguments);
  }
  
  /**
   * Auxiliary method generated to translate:
   * { if (populationSize < 1 || nDraws < 1 || maxRealization < 0) { return NEGATIVE_INFINITY } if (maxRealization + 1 < nDraws) { return NEGATIVE_INFINITY } if (populationSize <= maxRealization) { return NEGATIVE_INFINITY } return logBinomial(maxRealization, nDraws - 1) - logBinomial(populationSize, nDraws) }
   */
  private static double $generated__0(final IntVar maxRealization, final IntVar populationSize, final IntVar nDraws) {
    if (((((populationSize).intValue() < 1) || ((nDraws).intValue() < 1)) || ((maxRealization).intValue() < 0))) {
      return StaticUtils.NEGATIVE_INFINITY();
    }
    int _plus = BlangExtensions.operator_plus(maxRealization, 1);
    boolean _lessThan = (_plus < (nDraws).intValue());
    if (_lessThan) {
      return StaticUtils.NEGATIVE_INFINITY();
    }
    if (((populationSize).intValue() <= (maxRealization).intValue())) {
      return StaticUtils.NEGATIVE_INFINITY();
    }
    double _logBinomial = SpecialFunctions.logBinomial((maxRealization).intValue(), ((nDraws).intValue() - 1));
    double _logBinomial_1 = SpecialFunctions.logBinomial((populationSize).intValue(), (nDraws).intValue());
    return (_logBinomial - _logBinomial_1);
  }
  
  public static class $generated__0_class implements LogScaleFactor {
    public double logDensity() {
      return $generated__0(maxRealization, $generated__populationSize.get(), $generated__nDraws.get());
    }
    
    public String toString() {
      return "{ if (populationSize < 1 || nDraws < 1 || maxRealization < 0) { return NEGATIVE_INFINITY } if (maxRealization + 1 < nDraws) { return NEGATIVE_INFINITY } if (populationSize <= maxRealization) { return NEGATIVE_INFINITY } return logBinomial(maxRealization, nDraws - 1) - logBinomial(populationSize, nDraws) }";
    }
    
    private final IntVar maxRealization;
    
    private final Supplier<IntVar> $generated__populationSize;
    
    private final Supplier<IntVar> $generated__nDraws;
    
    public $generated__0_class(final IntVar maxRealization, final Supplier<IntVar> $generated__populationSize, final Supplier<IntVar> $generated__nDraws) {
      this.maxRealization = maxRealization;
      this.$generated__populationSize = $generated__populationSize;
      this.$generated__nDraws = $generated__nDraws;
    }
  }
  
  /**
   * Auxiliary method generated to translate:
   * { val List<Integer> list = (0 ..< populationSize).toList list.shuffle(rand) return list.subList(0, nDraws).max }
   */
  private static Integer $generated__1(final Random rand, final IntVar maxRealization, final IntVar populationSize, final IntVar nDraws) {
    final List<Integer> list = IterableExtensions.<Integer>toList(new ExclusiveRange(0, (populationSize).intValue(), true));
    Collections.shuffle(list, rand);
    return IterableExtensions.<Integer>max(list.subList(0, (nDraws).intValue()));
  }
  
  /**
   * Note: the generated code has the following properties used at runtime:
   *   - all arguments are annotated with a BlangVariable annotation
   *   - params additionally have a Param annotation
   *   - the order of the arguments is as follows:
   *     - first, all the random variables in the order they occur in the blang file
   *     - second, all the params in the order they occur in the blang file
   * 
   */
  public SerialSampling(@DeboxedName("maxRealization") final IntVar maxRealization, @Param @DeboxedName("populationSize") final Supplier<IntVar> $generated__populationSize, @Param @DeboxedName("nDraws") final Supplier<IntVar> $generated__nDraws) {
    this.maxRealization = maxRealization;
    this.$generated__populationSize = $generated__populationSize;
    this.$generated__nDraws = $generated__nDraws;
  }
  
  /**
   * A component can be either a distribution, support constraint, or another model  
   * which recursively defines additional components.
   */
  public Collection<ModelComponent> components() {
    ArrayList<ModelComponent> components = new ArrayList();
    
    { // Code generated by: (maxRealization; populationSize, nDraws) { if (populationSize < 1 || nDraws < 1 || maxRealization < 0) { return NEGATIVE_INFINITY } if (maxRealization + 1 < nDraws) { return NEGATIVE_INFINITY } if (populationSize <= maxRealization) { return NEGATIVE_INFINITY } return logBinomial(maxRealization, nDraws - 1) - logBinomial(populationSize, nDraws) }
      // Construction and addition of the factor/model:
      components.add(
        new $generated__0_class(maxRealization, $generated__populationSize, $generated__nDraws));
    }
    
    return components;
  }
  
  public void generate(final Random rand) {
    ((WritableIntVar) maxRealization).set(
      $generated__1(rand, maxRealization, $generated__populationSize.get(), $generated__nDraws.get())
    );
  }
  
  public IntVar realization() {
    return maxRealization;
  }
  
  /**
   * Returns an instance with fixed parameters values and conforming the Distribution interface. 
   * Useful when passing around distributions as parameters, e.g. for Dirichlet Process mixtures. 
   * 
   */
  public static IntDistribution distribution(@Param final IntVar populationSize, @Param final IntVar nDraws) {
    UnivariateModel<IntVar> univariateModel = new SerialSampling(
      new IntDistributionAdaptor.WritableIntVarImpl(), 
      new ConstantSupplier(populationSize), 
      new ConstantSupplier(nDraws)
    );
    Distribution<IntVar> distribution = new DistributionAdaptor(univariateModel);
    return new IntDistributionAdaptor(distribution);
  }
}
